import json

from lupa._lupa import LuaRuntime

from dredis.commands import run_command, SimpleString, CommandNotFound


class RedisScriptError(Exception):
    """Indicate error from calls to redis.call()"""


class RedisLua(object):

    def __init__(self, keyspace, lua_runtime):
        self._keyspace = keyspace
        self._lua_runtime = lua_runtime

    def call(self, cmd, *args):
        try:
            result = run_command(self._keyspace, cmd, args)
        except CommandNotFound:
            raise RedisScriptError('@user_script: Unknown Redis command called from Lua script')
        except Exception as exc:
            raise RedisScriptError(str(exc))
        else:
            return self._convert_redis_types_to_lua_types(result)

    def pcall(self, cmd, *args):
        try:
            return self.call(cmd, *args)
        except Exception as exc:
            table = self._lua_runtime.table()
            table['err'] = str(exc)
            return table

    def _convert_redis_types_to_lua_types(self, result):
        """
        Redis reply should be converted to the equivalent lua type
        The official implementation converts:
          * $-1 and *-1 to `false`
          * errors  to `{err=ERRORMSG}`
          * simple strings to `{ok=STRING}`
          * integers to numbers
          * arrays to lua tables following the previous conversions

        The implementation can be found at:
        https://github.com/antirez/redis/blob/5b4bec9d336655889641b134791dfdd2adc864cf/src/scripting.c#L106-L201
        """
        if isinstance(result, (tuple, list, set)):
            table = self._lua_runtime.table()
            for i, elem in enumerate(result, start=1):
                table[i] = self._convert_redis_types_to_lua_types(elem)
            return table
        elif result is None:
            return False
        elif result is True:
            return 1
        elif isinstance(result, SimpleString):
            table = self._lua_runtime.table()
            table["ok"] = result
            return table
        else:
            return result


class LuaRunner(object):
    def __init__(self, keyspace):
        self._runtime = LuaRuntime(unpack_returned_tuples=True)
        self._lua_table_type = type(self._runtime.table())
        self._redis_obj = RedisLua(keyspace, self._runtime)

    def run(self, script, keys, argv):
        self._runtime.execute('KEYS = {%s}' % ', '.join(map(json.dumps, keys)))
        self._runtime.execute('ARGV = {%s}' % ', '.join(map(json.dumps, argv)))
        script_function = self._runtime.eval('function(redis) {} end'.format(script))
        result = script_function(self._redis_obj)
        return self._convert_lua_types_to_redis_types(result)

    def _convert_lua_types_to_redis_types(self, result):
        def convert(value):
            """
            str -> str
            true -> 1
            false -> None
            number -> int
            table -> {
                if 'err' key is present, raise an error
                else if 'ok' key is present, return its value
                else convert to a list using the previous rules
            }

            Reference:
            https://github.com/antirez/redis/blob/5b4bec9d336655889641b134791dfdd2adc864cf/src/scripting.c#L273-L340

            """
            if isinstance(value, self._lua_table_type):
                if 'err' in value:
                    raise ValueError('ERR Error running script: {}'.format(value['err']))
                elif 'ok' in value:
                    return value['ok']
                else:
                    return map(convert, value.values())
            elif isinstance(value, (tuple, list, set)):
                return map(convert, value)
            elif value is True:
                return 1
            elif value is False:
                return None
            else:
                # assuming string at this point
                return value

        return convert(result)



class Parser(object):

    MAX_BUFSIZE = 1024 * 1024
    CRLF = '\r\n'

    def __init__(self, read_fn):
        self._buffer = ""
        self._read_fn = read_fn

    def _readline(self):
        if '\n' not in self._buffer:
            self._read_into_buffer()
        crlf_position = self._buffer.find(self.CRLF)
        result = self._buffer[:crlf_position]
        self._buffer = self._buffer[crlf_position + len(self.CRLF):]
        return result

    def _read_into_buffer(self, min_bytes=0):
        data = self._read_fn(self.MAX_BUFSIZE)
        self._buffer += data
        while data and len(self._buffer) < min_bytes:
            data = self._read_fn(self.MAX_BUFSIZE)
            self._buffer += data

    def _read(self, n_bytes):
        if len(self._buffer) < n_bytes:
            self._read_into_buffer(min_bytes=n_bytes)
        result = self._buffer[:n_bytes]
        self._buffer = self._buffer[n_bytes + 2:]
        return result

    def get_instructions(self):
        if not self._buffer:
            self._read_into_buffer()
        while self._buffer:
            instructions = self._readline()
            if not instructions:
                raise StopIteration()

            # the Redis protocol says that all commands are arrays, however,
            # Redis's own tests have commands like PING being sent as a Simple String
            if instructions.startswith('+'):
                yield instructions[1:].strip().split()
            # if instructions.startswith('*'):
            elif instructions.startswith('*'):
                # array of instructions
                array_length = int(instructions[1:])  # skip '*' char
                instruction_set = []
                for _ in range(array_length):
                    str_len = int(self._readline()[1:])  # skip '$' char
                    instruction = self._read(str_len)
                    instruction_set.append(instruction)
                yield instruction_set
            else:
                # inline instructions, saw them in the Redis tests
                for line in instructions.split('\r\n'):
                    yield line.strip().split()



import collections
import hashlib
import re

from dredis.lua import LuaRunner
from dredis.path import Path
from dredis.utils import to_float

DEFAULT_REDIS_DB = '0'
NUMBER_OF_REDIS_DATABASES = 15
DECIMAL_REGEX = re.compile('(\d+)\.0+$')


class DiskKeyspace(object):

    def __init__(self, root_dir):
        self._lua_runner = LuaRunner(self)
        self._root_directory = Path(root_dir)
        self._set_db_directory(DEFAULT_REDIS_DB)

    def _set_db_directory(self, db):
        self.directory = self._root_directory.join(db)

    def _key_path(self, key):
        return self.directory.join(key)

    def setup_directories(self):
        for db_id in range(NUMBER_OF_REDIS_DATABASES):
            self._root_directory.join(str(db_id)).makedirs(ignore_if_exists=True)

    def flushall(self):
        for db_id in range(NUMBER_OF_REDIS_DATABASES):
            self._root_directory.join(str(db_id)).reset()

    def flushdb(self):
        self.directory.reset()

    def select(self, db):
        self._set_db_directory(db)

    def incrby(self, key, increment=1):
        key_path = self._key_path(key)
        value_path = key_path.join('value')
        number = 0
        if self.exists(key):
            content = value_path.read()
            number = int(content)
        else:
            key_path.makedirs()
            self.write_type(key, 'string')
        result = number + increment
        value_path.write(str(result))
        return result

    def get(self, key):
        key_path = self._key_path(key)
        value_path = key_path.join('value')
        if self.exists(key):
            return value_path.read()
        else:
            return None

    def set(self, key, value):
        key_path = self._key_path(key)
        if not self.exists(key):
            key_path.makedirs()
            self.write_type(key, 'string')
        key_path.join('value').write(value)

    def getrange(self, key, start, end):
        value = self.get(key)
        if value is None:
            return ''
        else:
            if end < 0:
                end = len(value) + end
            end += 1  # inclusive
            return value[start:end]

    def sadd(self, key, value):
        key_path = self._key_path(key)
        values_path = key_path.join('values')
        if not self.exists(key):
            values_path.makedirs()
            self.write_type(key, 'set')
        fname = self._get_filename_hash(value)
        value_path = values_path.join(fname)
        if value_path.exists():
            return 0
        else:
            value_path.write(value)
            return 1

    def smembers(self, key):
        result = set()
        if self.exists(key):
            key_path = self._key_path(key)
            values_path = key_path.join('values')
            for fname in values_path.listdir():
                content = values_path.join(fname).read()
                result.add(content)
        return result

    def sismember(self, key, value):
        key_path = self._key_path(key)
        values_path = key_path.join('values')
        fname = self._get_filename_hash(value)
        value_path = values_path.join(fname)
        return value_path.exists()

    def scard(self, key):
        return len(self.smembers(key))

    def delete(self, *keys):
        result = 0
        for key in keys:
            if self.exists(key):
                key_path = self._key_path(key)
                key_path.delete()
                result += 1
        return result

    def zadd(self, key, score, value):
        """
        /path/scores/10 -> '"x1"\n"x2"'
        /path/scores/20 -> '"y"'
        /path/scores/30 -> '"z"'
        /path/values/hash("x1") -> "10"
        """

        # if `score` has 0 as the decimal point, trim it: 10.00 -> 10
        match = DECIMAL_REGEX.match(score)
        if match:
            score = match.group(1)

        key_path = self._key_path(key)
        scores_path = key_path.join('scores')
        values_path = key_path.join('values')
        if not key_path.exists():
            scores_path.makedirs()
            values_path.makedirs()
            self.write_type(key, 'zset')

        score_path = scores_path.join(score)
        value_path = values_path.join(self._get_filename_hash(value))
        if value_path.exists():
            result = 0
            previous_score = value_path.read()
            if previous_score == score:
                return result
            else:
                previous_score_path = scores_path.join(previous_score)
                previous_score_path.remove_line(value)
        else:
            result = 1

        value_path.write(score)
        score_path.append(value)
        return result

    def write_type(self, key, name):
        key_path = self._key_path(key)
        type_path = key_path.join('type')
        type_path.write(name)

    def zrange(self, key, start, stop, with_scores):
        key_path = self._key_path(key)
        lines = []
        if with_scores:
            n_elems = 2
        else:
            n_elems = 1
        scores_path = key_path.join('scores')
        if scores_path.exists():
            scores = sorted(scores_path.listdir(), key=float)
            for score in scores:
                sublist = sorted(scores_path.join(score).readlines())
                for line in sublist:
                    lines.append(line)
                    if with_scores:
                        lines.append(score)
        if stop < 0:
            end = len(lines) + stop * n_elems + n_elems
        else:
            end = (stop + 1) * n_elems

        if start < 0:
            begin = max(0, len(lines) + start * n_elems)
        else:
            begin = start * n_elems

        return lines[begin:end]

    def zcard(self, key):
        key_path = self._key_path(key)
        values_path = key_path.join('values')
        if values_path.exists():
            return len(values_path.listdir())
        else:
            return 0

    def zscore(self, key, member):
        key_path = self._key_path(key)
        value_path = key_path.join('values').join(self._get_filename_hash(member))
        if value_path.exists():
            return value_path.read().strip()
        else:
            return None

    def eval(self, script, keys, argv):
        return self._lua_runner.run(script, keys, argv)

    def zrem(self, key, *members):
        """
        /path/scores/10 -> "x1\nx2"
        /path/scores/20 -> "y"
        /path/scores/30 -> "z"
        /path/values/hash(x) -> 1
        """
        key_path = self._key_path(key)
        scores_path = key_path.join('scores')
        values_path = key_path.join('values')
        result = 0
        for member in members:
            value_path = values_path.join(self._get_filename_hash(member))
            if not value_path.exists():
                continue
            result += 1
            score = value_path.read().strip()
            score_path = scores_path.join(score)
            score_path.remove_line(member)
            value_path.delete()
        # empty zset should be removed from keyspace
        if scores_path.empty_directory():
            self.delete(key)
        return result

    def zrangebyscore(self, key, min_score, max_score, withscores=False, offset=0, count=float('+inf')):
        result = []
        num_elems_read = 0
        if withscores:
            num_elems_per_entry = 2
        else:
            num_elems_per_entry = 1
        key_path = self._key_path(key)
        scores_path = key_path.join('scores')
        if scores_path.exists():
            scores = sorted(scores_path.listdir(), key=float)
            score_range = ScoreRange(min_score, max_score)
            scores = [score for score in scores if score_range.check(float(score))]
            for score in scores:
                lines = sorted(scores_path.join(score).readlines())
                for line in lines:
                    num_elems_read += 1
                    if len(result) / num_elems_per_entry >= count:
                        return result
                    if num_elems_read > offset:
                        result.append(line)
                        if withscores:
                            result.append(str(score))
        return result

    def zcount(self, key, min_score, max_score):
        return len(self.zrangebyscore(key, min_score, max_score))

    def zrank(self, key, member):
        key_path = self._key_path(key)
        value_path = key_path.join('values').join(self._get_filename_hash(member))
        if value_path.exists():
            scores_path = key_path.join('scores')
            scores = sorted(scores_path.listdir(), key=float)
            member_score = value_path.read().strip()
            rank = 0
            for score in scores:
                score_path = scores_path.join(str(score))
                lines = score_path.readlines()  # FIXME: move this to be inside the `if` block
                if score == member_score:
                    for line in lines:
                        if line == member:
                            return rank
                        else:
                            rank += 1
                else:
                    rank += len(lines)

            return rank
        else:
            return None

    def zunionstore(self, destination, keys, weights):
        union = collections.defaultdict(list)
        for (key, weight) in zip(keys, weights):
            elem_with_scores = self.zrange(key, 0, -1, with_scores=True)
            while elem_with_scores:
                member = elem_with_scores.pop(0)
                score = elem_with_scores.pop(0)
                union[member].append(float(score) * weight)
        aggregate_fn = sum  # FIXME: redis also supports MIN and MAX

        result = 0
        for member, scores in union.items():
            score = aggregate_fn(scores)
            result += self.zadd(destination, str(score), member)
        return result

    def type(self, key):
        key_path = self._key_path(key)
        if key_path.exists():
            type_path = key_path.join('type')
            return type_path.read()
        else:
            return 'none'

    def keys(self, pattern):
        return self.directory.listdir(pattern)

    def dbsize(self):
        return len(self.directory.listdir())

    def exists(self, *keys):
        result = 0
        for key in keys:
            key_path = self._key_path(key)
            if key_path.exists():
                result += 1
        return result

    def hset(self, key, field, value):
        key_path = self._key_path(key)
        fields_path = key_path.join('fields')
        if not self.exists(key):
            fields_path.makedirs()
        field_path = fields_path.join(field)
        if field_path.exists():
            result = 0
        else:
            result = 1
        field_path.write(value)
        self.write_type(key, 'hash')
        return result

    def hsetnx(self, key, field, value):
        key_path = self._key_path(key)
        fields_path = key_path.join('fields')
        if not self.exists(key):
            fields_path.makedirs()
        field_path = fields_path.join(field)
        result = 0
        # only set if not set before
        if not field_path.exists():
            result = 1
            field_path.write(value)
        self.write_type(key, 'hash')
        return result

    def hdel(self, key, *fields):
        result = 0
        key_path = self._key_path(key)
        fields_path = key_path.join('fields')
        for field in fields:
            field_path = fields_path.join(field)
            if field_path.exists():
                field_path.delete()
                result += 1
        # remove empty hashes from keyspace
        if fields_path.empty_directory():
            self.delete(key)
        return result

    def hget(self, key, field):
        key_path = self._key_path(key)
        fields_path = key_path.join('fields')
        field_path = fields_path.join(field)
        if field_path.exists():
            result = field_path.read()
        else:
            result = None
        return result

    def hkeys(self, key):
        key_path = self._key_path(key)
        fields_path = key_path.join('fields')
        if fields_path.exists():
            result = fields_path.listdir()
        else:
            result = []
        return result

    def hvals(self, key):
        result = []
        key_path = self._key_path(key)
        fields_path = key_path.join('fields')
        if fields_path.exists():
            for field in fields_path.listdir():
                field_path = fields_path.join(field)
                value = field_path.read()
                result.append(value)
        return result

    def hlen(self, key):
        key_path = self._key_path(key)
        fields_path = key_path.join('fields')
        result = 0
        if fields_path.exists():
            result = len(fields_path.listdir())
        return result

    def hincrby(self, key, field, increment):
        before = self.hget(key, field) or '0'
        new_value = int(before) + int(increment)
        self.hset(key, field, str(new_value))
        return new_value

    def hgetall(self, key):
        keys = self.hkeys(key)
        values = self.hvals(key)
        result = []
        for (k, v) in zip(keys, values):
            result.append(k)
            result.append(v)
        return result

    def _get_filename_hash(self, value):
        return hashlib.md5(value).hexdigest()


class ScoreRange(object):

    def __init__(self, min_value, max_value):
        self._min_value = min_value
        self._max_value = max_value

    def check(self, value):
        if self._min_value.startswith('('):
            if to_float(self._min_value[1:]) >= value:
                return False
        elif to_float(self._min_value) > value:
            return False

        if self._max_value.startswith('('):
            if to_float(self._max_value[1:]) <= value:
                return False
        elif to_float(self._max_value) < value:
            return False

        return True



import argparse
import asyncore
import json
import logging
import os.path
import socket
import tempfile
import traceback

import sys

from dredis import __version__
from dredis.commands import run_command, SimpleString, CommandNotFound
from dredis.keyspace import DiskKeyspace
from dredis.lua import RedisScriptError
from dredis.parser import Parser


logger = logging.getLogger('dredis')

KEYSPACES = {}
ROOT_DIR = None  # defined by `main()`


def not_found(send_fn, cmd):
    err(send_fn, "unknown command '{}'".format(cmd))


def err(send_fn, msg):
    send_fn("-ERR {}\r\n".format(msg))


def error(send_fn, msg):
    send_fn('-{}\r\n'.format(msg))


def execute_cmd(keyspace, send_fn, cmd, *args):
    try:
        result = run_command(keyspace, cmd, args)
    except (ValueError, RedisScriptError) as exc:
        error(send_fn, str(exc))
    except CommandNotFound:
        not_found(send_fn, cmd)
    except SyntaxError as exc:
        err(send_fn, str(exc))
    except Exception:
        err(send_fn, json.dumps(traceback.format_exc()))
    else:
        transmit(send_fn, result)


def transmit(send_fn, result):
    if result is None:
        send_fn('$-1\r\n')
    elif isinstance(result, int):
        send_fn(':{}\r\n'.format(result))
    elif isinstance(result, SimpleString):
        send_fn('+{}\r\n'.format(result))
    elif isinstance(result, basestring):
        send_fn('${}\r\n{}\r\n'.format(len(result), result))
    elif isinstance(result, (set, list, tuple)):
        send_fn('*{}\r\n'.format(len(result)))
        for element in result:
            transmit(send_fn, element)
    else:
        assert False, 'couldnt catch a response for {} (type {})'.format(repr(result), type(result))


class CommandHandler(asyncore.dispatcher):

    def handle_read(self):
        parser = Parser(self.recv)
        for cmd in parser.get_instructions():
            logger.debug('{} data = {}'.format(self.addr, repr(cmd)))
            execute_cmd(self.keyspace, self.debug_send, *cmd)

    def debug_send(self, *args):
        logger.debug("out={}".format(repr(args)))
        return self.send(*args)

    def handle_close(self):
        logger.debug("closing {}".format(self.addr))
        self.close()
        if self.addr in KEYSPACES:
            del KEYSPACES[self.addr]

    @property
    def keyspace(self):
        if self.addr not in KEYSPACES:
            KEYSPACES[self.addr] = DiskKeyspace(ROOT_DIR)
        return KEYSPACES[self.addr]


class RedisServer(asyncore.dispatcher):

    def __init__(self, host, port):
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((host, port))
        self.listen(1024)

    def handle_accept(self):
        pair = self.accept()
        if pair is not None:
            sock, addr = pair
            logger.debug('Incoming connection from %s' % repr(addr))
            CommandHandler(sock)


def setup_logging(level):
    logger.setLevel(level)
    formatter = logging.Formatter('%(asctime)s [%(levelname)s] %(message)s')
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


def main():
    parser = argparse.ArgumentParser(version=__version__)
    parser.add_argument('--host', default='127.0.0.1', help='server host (defaults to %(default)s)')
    parser.add_argument('--port', default='6377', type=int, help='server port (defaults to %(default)s)')
    parser.add_argument('--dir', default=None,
                        help='directory to save data (defaults to a temporary directory)')
    parser.add_argument('--debug', action='store_true', help='enable debug logs')
    parser.add_argument('--flushall', action='store_true', default=False, help='run FLUSHALL on startup')
    args = parser.parse_args()

    global ROOT_DIR
    if args.dir:
        ROOT_DIR = args.dir
    else:
        ROOT_DIR = tempfile.mkdtemp(prefix="redis-test-")

    if args.debug:
        setup_logging(logging.DEBUG)
    else:
        setup_logging(logging.INFO)

    keyspace = DiskKeyspace(ROOT_DIR)
    if args.flushall:
        keyspace.flushall()
    else:
        keyspace.setup_directories()

    RedisServer(args.host, args.port)

    logger.info("Port: {}".format(args.port))
    logger.info("Root directory: {}".format(ROOT_DIR))
    logger.info('PID: {}'.format(os.getpid()))
    logger.info('Ready to accept connections')

    try:
        asyncore.loop(use_poll=True)
    except KeyboardInterrupt:
        logger.info("Shutting down...")


if __name__ == '__main__':
    main()



import logging
from functools import wraps

from dredis.utils import to_float

logger = logging.getLogger(__name__)


REDIS_COMMANDS = {}
SYNTAXERR = SyntaxError('syntax error')


def _check_arity(expected_arity, passed_arity, cmd_name):
    syntax_err = SyntaxError("wrong number of arguments for '{}' command".format(cmd_name.lower()))
    if expected_arity < 0:  # minimum arity
        if passed_arity < -expected_arity:
            raise syntax_err
    elif expected_arity > 0:  # exact match
        if passed_arity != expected_arity:  # exact arity
            raise syntax_err
    else:  # ignore
        # ignore arity of 0 (at the moment it's only for `COMMAND`).
        # it could be set to 1 but the redis source has it as 0, just following their implementation.
        # source: https://github.com/antirez/redis/blob/cb51bb4320d2240001e8fc4a522d59fb28259703/src/server.c#L296
        return


def command(cmd_name, arity):
    def decorator(fn):
        @wraps(fn)
        def newfn(keyspace, *args, **kwargs):
            # redis includes the command name in the arity, thus adding 1
            passed_arity = 1 + len(args) + len(kwargs)
            _check_arity(arity, passed_arity, cmd_name)
            return fn(keyspace, *args, **kwargs)
        newfn.arity = arity
        REDIS_COMMANDS[cmd_name] = newfn
        return newfn
    return decorator


class SimpleString(str):
    pass


"""
*******************
* Server commands *
*******************
"""


@command('COMMAND', arity=0)
def cmd_command(keyspace):
    result = []
    for cmd in REDIS_COMMANDS:
        result.append(cmd.upper())
    return result


@command('FLUSHALL', arity=-1)
def cmd_flushall(keyspace, *args):
    # TODO: we don't support ASYNC flushes
    keyspace.flushall()
    return SimpleString('OK')


@command('FLUSHDB', arity=-1)
def cmd_flushdb(keyspace, *args):
    # TODO: we don't support ASYNC flushes
    keyspace.flushdb()
    return SimpleString('OK')


@command('DBSIZE', arity=1)
def cmd_dbsize(keyspace):
    return keyspace.dbsize()


"""
****************
* Key commands *
****************
"""


@command('DEL', arity=-2)
def cmd_del(keyspace, *keys):
    return keyspace.delete(*keys)


@command('TYPE', arity=2)
def cmd_type(keyspace, key):
    return keyspace.type(key)


@command('KEYS', arity=2)
def cmd_keys(keyspace, pattern):
    return keyspace.keys(pattern)


@command('EXISTS', arity=-2)
def cmd_exists(keyspace, *keys):
    return keyspace.exists(*keys)


"""
***********************
* Connection commands *
***********************
"""


@command('PING', arity=-1)
def cmd_ping(keyspace, message=SimpleString('PONG')):
    return message


@command('SELECT', arity=2)
def cmd_select(keyspace, db):
    keyspace.select(db)
    return SimpleString('OK')


"""
*******************
* String commands *
*******************
"""


@command('SET', arity=-3)
def cmd_set(keyspace, key, value, *args):
    if len(args):
        raise SyntaxError('No support for EX|PX and NX|XX at the moment.')
    keyspace.set(key, value)
    return SimpleString('OK')


@command('GET', arity=2)
def cmd_get(keyspace, key):
    return keyspace.get(key)


@command('INCR', arity=2)
def cmd_incr(keyspace, key):
    return keyspace.incrby(key, 1)


@command('INCRBY', arity=3)
def cmd_incrby(keyspace, key, increment):
    return keyspace.incrby(key, int(increment))


@command('GETRANGE', arity=4)
def cmd_getrange(keyspace, key, start, end):
    return keyspace.getrange(key, int(start), int(end))


"""
****************
* Set commands *
****************
"""


@command('SADD', arity=-3)
def cmd_sadd(keyspace, key, *values):
    count = 0
    for value in values:
        count += keyspace.sadd(key, value)
    return count


@command('SMEMBERS', arity=2)
def cmd_smembers(keyspace, key):
    return keyspace.smembers(key)


@command('SCARD', arity=2)
def cmd_scard(keyspace, key):
    return keyspace.scard(key)


@command('SISMEMBER', arity=3)
def cmd_sismember(keyspace, key, value):
    return int(keyspace.sismember(key, value))


"""
**********************
* Scripting commands *
**********************
"""


@command('EVAL', arity=-3)
def cmd_eval(keyspace, script, numkeys, *args):
    numkeys = int(numkeys)
    keys = args[:numkeys]
    argv = args[numkeys:]
    return keyspace.eval(script, keys, argv)


"""
***********************
* Sorted set commands *
***********************
"""


@command('ZADD', arity=-4)
def cmd_zadd(keyspace, key, *flat_pairs):
    if len(flat_pairs) % 2 != 0:
        raise SYNTAXERR

    count = 0
    pairs = zip(flat_pairs[0::2], flat_pairs[1::2])  # [1, 2, 3, 4] -> [(1,2), (3,4)]
    for score, value in pairs:
        _validate_zset_score(score)
        count += keyspace.zadd(key, score, value)
    return count


@command('ZRANGE', arity=-4)
def cmd_zrange(keyspace, key, start, stop, *args):
    with_scores = False
    if args:
        if args[0].lower() == 'withscores':
            with_scores = True
        else:
            raise SYNTAXERR
    return keyspace.zrange(key, int(start), int(stop), with_scores)


@command('ZCARD', arity=2)
def cmd_zcard(keyspace, key):
    return keyspace.zcard(key)


@command('ZREM', arity=-3)
def cmd_zrem(keyspace, key, *members):
    return keyspace.zrem(key, *members)


@command('ZSCORE', arity=3)
def cmd_zscore(keyspace, key, member):
    return keyspace.zscore(key, member)


@command('ZRANK', arity=3)
def cmd_zrank(keyspace, key, member):
    return keyspace.zrank(key, member)


@command('ZCOUNT', arity=4)
def cmd_zcount(keyspace, key, min_score, max_score):
    return keyspace.zcount(key, min_score, max_score)


@command('ZRANGEBYSCORE', arity=-4)
def cmd_zrangebyscore(keyspace, key, min_score, max_score, *args):
    withscores = False
    offset = 0
    count = float('+inf')
    args = list(args)
    while args:
        arg = args.pop(0)
        if len(args) >= 0 and arg.lower() == 'withscores':
            withscores = True
        elif len(args) >= 2 and arg.lower() == 'limit':
            offset = int(args.pop(0))
            count = int(args.pop(0))
        else:
            raise SYNTAXERR

    _validate_zset_score(min_score)
    _validate_zset_score(max_score)

    members = keyspace.zrangebyscore(
        key, min_score, max_score, withscores=withscores, offset=offset, count=count)
    return members


def _validate_zset_score(score):
    clean_score = score.strip('(').lower().replace('nan', 'invalid')
    try:
        to_float(clean_score)
    except ValueError:
        raise SyntaxError("min or max is not a float")


@command('ZUNIONSTORE', arity=-4)
def cmd_zunionstore(keyspace, destination, numkeys, *args):
    keys = []
    weights = []
    args = list(args)
    while args:
        arg = args.pop(0)
        if arg == 'WEIGHTS':
            weights = map(float, args)
            break
        else:
            keys.append(arg)
    if len(weights) > len(keys):
        raise SYNTAXERR
    ones = [1] * (len(keys) - len(weights))  # fill in with default weight of 1
    weights = weights + ones
    return keyspace.zunionstore(destination, keys, weights)


"""
*******************
* Hash commands *
*******************
"""


@command('HSET', arity=-4)
def cmd_hset(keyspace, key, *pairs):
    if len(pairs) % 2 != 0:
        # HSET is going to replace HMSET,
        # see https://github.com/antirez/redis/pull/5334#issuecomment-419194180 for more details
        raise SyntaxError('wrong number of arguments for HMSET')
    count = 0
    for field, value in zip(pairs[0::2], pairs[1::2]):
        count += keyspace.hset(key, field, value)
    return count


@command('HDEL', arity=-3)
def cmd_hdel(keyspace, key, *fields):
    return keyspace.hdel(key, *fields)


@command('HSETNX', arity=4)
def cmd_hsetnx(keyspace, key, field, value):
    return keyspace.hsetnx(key, field, value)


@command('HGET', arity=3)
def cmd_hget(keyspace, key, value):
    return keyspace.hget(key, value)


@command('HKEYS', arity=2)
def cmd_hkeys(keyspace, key):
    return keyspace.hkeys(key)


@command('HVALS', arity=2)
def cmd_hvals(keyspace, key):
    return keyspace.hvals(key)


@command('HLEN', arity=2)
def cmd_hlen(keyspace, key):
    return keyspace.hlen(key)


@command('HINCRBY', arity=4)
def cmd_hincrby(keyspace, key, field, increment):
    return keyspace.hincrby(key, field, increment)


@command('HGETALL', arity=2)
def cmd_hgetall(keyspace, key):
    return keyspace.hgetall(key)


class CommandNotFound(Exception):
    """Exception to flag not found Redis command"""


def run_command(keyspace, cmd, args):
    logger.debug('[run_command] cmd={}, args={}'.format(repr(cmd), repr(args)))

    str_args = map(str, args)
    if cmd.upper() not in REDIS_COMMANDS:
        raise CommandNotFound()
    else:
        return REDIS_COMMANDS[cmd.upper()](keyspace, *str_args)



import errno
import fnmatch
import json
import os.path
import shutil
import six
import sys


class Path(str):

    def join(self, path):
        return Path(os.path.join(self, path))

    def reset(self):
        shutil.rmtree(self, ignore_errors=True)
        os.makedirs(self)

    def read(self):
        with open(self, 'r') as f:
            result = f.read()
        return self._deserialize(result)

    def write(self, content):
        with open(self, 'w') as f:
            f.write(self._serialize(content))

    def delete(self):
        if os.path.isfile(self):
            os.remove(self)
        else:
            shutil.rmtree(self)

    def append(self, line):
        with open(self, 'a') as f:
            f.write(self._serialize(line) + '\n')

    def readlines(self):
        with open(self) as f:
            lines = f.readlines()
        return map(self._deserialize, lines)

    def exists(self):
        return os.path.exists(self)

    def makedirs(self, ignore_if_exists=False):
        try:
            return os.makedirs(self)
        except OSError as exc:
            if ignore_if_exists and exc.errno == errno.EEXIST:
                pass
            else:
                six.reraise(*sys.exc_info())

    def listdir(self, pattern=None):
        all_files = os.listdir(self)
        if pattern is None:
            return all_files
        else:
            return list(fnmatch.filter(all_files, pattern))

    def remove_line(self, line_to_remove):
        lines = self.readlines()
        if line_to_remove in lines:
            lines.remove(line_to_remove)
            if len(lines) == 0:
                os.remove(self)
            else:
                new_content = "".join(self._serialize(line) + "\n" for line in lines)
                with open(self, 'w') as f:
                    f.write(new_content)

    def empty_directory(self):
        return self.exists() and not self.listdir()

    def _deserialize(self, value):
        return json.loads(value)

    def _serialize(self, value):
        return json.dumps(value)
